/********************************************************************************

Copyright 2019  Thomas Fischer <fischer@unix-ag.uni-kl.de> and others

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

********************************************************************************/

#ifndef KBIBTEX_FILES_TEST_RAWDATA_H
#define KBIBTEX_FILES_TEST_RAWDATA_H

static const int bug19489NumElements = 1;
static const int bug19489NumEntries = 1;
static const QString bug19489LastEntryId = QStringLiteral("bart:04:1242");
static const QString bug19489LastAuthor = QStringLiteral("Ralph");
static const QByteArray bug19489LastAuthors = QByteArray::fromHex("1be83d735510848b88a08e43c2e114d9");
static const QByteArray bug19489FilesUrlsDoi = QByteArray::fromHex("544dc57e2a9908cfb35f35152462ac2e");
static const int nameswithbracesNumElements = 1;
static const int nameswithbracesNumEntries = 1;
static const QString nameswithbracesLastEntryId = QStringLiteral("names1");
static const QString nameswithbracesLastAuthor = QStringLiteral("{{{{{LastName3A LastName3B}}}}}");
static const QByteArray nameswithbracesLastAuthors = QByteArray::fromHex("c7f3a5aaf7b68ac8774fab52c3b1b29d");
static const QByteArray nameswithbracesFilesUrlsDoi = QByteArray::fromHex("d41d8cd98f00b204e9800998ecf8427e");
static const int duplicatesNumElements = 23;
static const int duplicatesNumEntries = 23;
static const QString duplicatesLastEntryId = QStringLiteral("books/aw/Sedgewick88");
static const QString duplicatesLastAuthor = QStringLiteral("Sedgewick");
static const QByteArray duplicatesLastAuthors = QByteArray::fromHex("ab50e9b4ced3aac2060f9a4fea2c6f5d");
static const QByteArray duplicatesFilesUrlsDoi = QByteArray::fromHex("b8a7d0f38c03ada4615989c626a77a1f");
static const int minixNumElements = 163;
static const int minixNumEntries = 123;
static const QString minixLastEntryId = QStringLiteral("Jesshope:2006:ACS");
static const QString minixLastAuthor = QStringLiteral("Egan");
static const QByteArray minixLastAuthors = QByteArray::fromHex("9f58143bfbe15014aee1e96df97aff6d");
static const QByteArray minixFilesUrlsDoi = QByteArray::fromHex("4e64d5806c9be0cf2d77a152157d044c");
static const int bug19484refsNumElements = 641;
static const int bug19484refsNumEntries = 641;
static const QString bug19484refsLastEntryId = QStringLiteral("Bagnara-etal-2002");
static const QString bug19484refsLastAuthor = QStringLiteral("Hill");
static const QByteArray bug19484refsLastAuthors = QByteArray::fromHex("08b9d2222e023290cd9db07969836ff2");
static const QByteArray bug19484refsFilesUrlsDoi = QByteArray::fromHex("71a66fb634275a6f46e5e0828fbceb83");
static const int bug19362file15701databaseNumElements = 911;
static const int bug19362file15701databaseNumEntries = 911;
static const QString bug19362file15701databaseLastEntryId = QStringLiteral("New1");
static const QString bug19362file15701databaseLastAuthor = QStringLiteral("Sunder");
static const QByteArray bug19362file15701databaseLastAuthors = QByteArray::fromHex("68beb5da6ddf13c110ab53c7c8b262db");
static const QByteArray bug19362file15701databaseFilesUrlsDoi = QByteArray::fromHex("0ac293b3abbfa56867e5514d8bb68614");
static const int digiplayNumElements = 3074;
static const int digiplayNumEntries = 3074;
static const QString digiplayLastEntryId = QStringLiteral("1180");
static const QString digiplayLastAuthor = QStringLiteral("Huizinga");
static const QByteArray digiplayLastAuthors = QByteArray::fromHex("1119a0d598db77a9e956ee21d12b2580");
static const QByteArray digiplayFilesUrlsDoi = QByteArray::fromHex("5d5bf8178652107ab160bc697b5b008f");
static const int backslashNumElements = 1;
static const int backslashNumEntries = 1;
static const QString backslashLastEntryId = QStringLiteral("backslash-test");
static const QString backslashLastAuthor = QStringLiteral("Doe");
static const QByteArray backslashLastAuthors = QByteArray::fromHex("ad695f53ae7569fb981fc95598e27e67");
static const QByteArray backslashFilesUrlsDoi = QByteArray::fromHex("f9f35d6b95b0676751bc613d1d60aa6b");
static const int bug379443attachment105313IOPEXPORTBIBNumElements = 1;
static const int bug379443attachment105313IOPEXPORTBIBNumEntries = 1;
static const QString bug379443attachment105313IOPEXPORTBIBLastEntryId = QStringLiteral("1748-0221-3-08-S08004");
static const QString bug379443attachment105313IOPEXPORTBIBLastAuthor = QStringLiteral("Yuldashev");
static const QByteArray bug379443attachment105313IOPEXPORTBIBLastAuthors = QByteArray::fromHex("31c8f8adee315b4c418da3899c7ee3b6");
static const QByteArray bug379443attachment105313IOPEXPORTBIBFilesUrlsDoi = QByteArray::fromHex("686d8f0ae0a1d66e506235b09486dd37");
static const int bug21870politoNumElements = 736;
static const int bug21870politoNumEntries = 721;
static const QString bug21870politoLastEntryId = QStringLiteral("BusseNiehrWengeler:Brisante05");
static const QString bug21870politoLastAuthor = QStringLiteral("Wengeler");
static const QByteArray bug21870politoLastAuthors = QByteArray::fromHex("cfde978f186b14ff9594295a4980d3c5");
static const QByteArray bug21870politoFilesUrlsDoi = QByteArray::fromHex("ac7d960f7978710fcee16e975e7c1960");
static const int cloudduplicatesNumElements = 21;
static const int cloudduplicatesNumEntries = 21;
static const QString cloudduplicatesLastEntryId = QStringLiteral("rao2012survey");
static const QString cloudduplicatesLastAuthor = QStringLiteral("Reddy");
static const QByteArray cloudduplicatesLastAuthors = QByteArray::fromHex("9ed190c716f7a3d7e472a88177ea0d7a");
static const QByteArray cloudduplicatesFilesUrlsDoi = QByteArray::fromHex("7958532962bc1528903679238aee830e");

#endif // KBIBTEX_FILES_TEST_RAWDATA_H
